@@ -13,6 +13,8 @@
 # limitations under the License.
 # =============================================================================
 """Implements the yolov2 loss."""
+import math
+
 import tensorflow as tf
 
 
@@ -25,7 +27,8 @@ def yolo_loss(y_true,
               lambda_class=1.0,
               iou_threshold=0.4,
               step=0,
-              is_training=True):
+              is_training=True,
+              version="vanilla"):
     """Defines the yolov2 multipart loss function.
 
     Args:
@@ -59,14 +62,19 @@ def yolo_loss(y_true,
         y_true)
 
     # Calculate coordinate loss_xywh.
-    loss_xywh = _loss_xywh(
-        true_box_xy,
-        p_box_xy,
-        true_box_wh,
-        p_box_wh,
-        true_box_conf,
-        lambda_coord,
-    )
+    loss_xywh = None
+    if version == "vanilla":
+        loss_xywh = _loss_xywh(
+            true_box_xy,
+            p_box_xy,
+            true_box_wh,
+            p_box_wh,
+            true_box_conf,
+            lambda_coord,
+        )
+    elif version == "ciou":
+        loss_xywh = _cIoU_loss(true_box_xy, p_box_xy, true_box_wh, p_box_wh,
+                               true_box_conf)
 
     # Calculate loss for the class probabilities.
     loss_class = _loss_class(
@@ -268,6 +276,63 @@ def _loss_xywh(true_box_xy, pred_box_xy, true_box_wh, pred_box_wh,
     return loss_xywh
 
 
+def _cIoU_loss(true_box_xy, pred_box_xy, true_box_wh, pred_box_wh,
+               true_box_conf):
+    """Computes the CIoU Localization Loss
+    Inspired by: https://github.com/robbebluecp/tf2-yolov4/blob/online/loss.py
+    """
+    b1_xy = true_box_xy
+    b1_wh = true_box_wh
+    b1_wh_half = b1_wh / 2.
+    b1_mins = b1_xy - b1_wh_half
+    b1_maxes = b1_xy + b1_wh_half
+
+    b2_xy = pred_box_xy
+    b2_wh = pred_box_wh
+    b2_wh_half = b2_wh / 2.
+    b2_mins = b2_xy - b2_wh_half
+    b2_maxes = b2_xy + b2_wh_half
+
+    intersect_mins = tf.keras.backend.maximum(b1_mins, b2_mins)
+    intersect_maxes = tf.keras.backend.minimum(b1_maxes, b2_maxes)
+    intersect_wh = tf.keras.backend.maximum(intersect_maxes - intersect_mins,
+                                            0.)
+    intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]
+    b1_area = b1_wh[..., 0] * b1_wh[..., 1]
+    b2_area = b2_wh[..., 0] * b2_wh[..., 1]
+    union_area = b1_area + b2_area - intersect_area
+    # calculate IoU, add epsilon in denominator to avoid dividing by 0
+    iou = intersect_area / (union_area + tf.keras.backend.epsilon())
+
+    # box center distance
+    center_distance = tf.keras.backend.sum(tf.keras.backend.square(b1_xy -
+                                                                   b2_xy),
+                                           axis=-1)
+    # get enclosed area
+    enclose_mins = tf.keras.backend.minimum(b1_mins, b2_mins)
+    enclose_maxes = tf.keras.backend.maximum(b1_maxes, b2_maxes)
+    enclose_wh = tf.keras.backend.maximum(enclose_maxes - enclose_mins, 0.0)
+    # get enclosed diagonal distance
+    enclose_diagonal = tf.keras.backend.sum(
+        tf.keras.backend.square(enclose_wh), axis=-1)
+    # calculate DIoU, add epsilon in denominator to avoid dividing by 0
+    diou = 1 - iou + center_distance / \
+        (enclose_diagonal + tf.keras.backend.epsilon())
+
+    v = 4 * tf.keras.backend.square(
+        tf.math.atan2(b1_wh[..., 0], b1_wh[..., 1]) -
+        tf.math.atan2(b2_wh[..., 0], b2_wh[..., 1])) / (math.pi * math.pi)
+    alpha = v / (1.0 - iou + v)
+    ciou = diou + alpha * v
+
+    ciou = tf.expand_dims(ciou, -1)
+    ciou = tf.cast(ciou, dtype=tf.float32)
+    coord_mask = tf.cast(true_box_conf, dtype=tf.float32)
+
+    ciou = tf.keras.backend.sum(ciou * coord_mask)
+    return ciou
+
+
 def _loss_class(
         true_box_class,
         p_box_class,
